from .icp import *

# TODO: move contents from nonconformist.icp here


# -----------------------------------------------------------------------------
# TcpClassifier
# -----------------------------------------------------------------------------
class TcpClassifier(BaseEstimator, ClassifierMixin):
    """Transductive conformal classifier.

    Parameters
    ----------
    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity of
            calibration examples and test patterns. Should implement ``fit(x, y)``
            and ``calc_nc(x, y)``.

    smoothing : boolean
            Decides whether to use stochastic smoothing of p-values.

    Attributes
    ----------
    train_x : numpy array of shape [n_cal_examples, n_features]
            Inputs of training set.

    train_y : numpy array of shape [n_cal_examples]
            Outputs of calibration set.

    nc_function : BaseScorer
            Nonconformity scorer object used to calculate nonconformity scores.

    classes : numpy array of shape [n_classes]
            List of class labels, with indices corresponding to output columns
             of TcpClassifier.predict()

    See also
    --------
    IcpClassifier

    References
    ----------
    .. [1] Vovk, V., Gammerman, A., & Shafer, G. (2005). Algorithmic learning
    in a random world. Springer Science & Business Media.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.svm import SVC
    >>> from nonconformist.base import ClassifierAdapter
    >>> from nonconformist.cp import TcpClassifier
    >>> from nonconformist.nc import ClassifierNc, MarginErrFunc
    >>> iris = load_iris()
    >>> idx = np.random.permutation(iris.target.size)
    >>> train = idx[:int(idx.size / 2)]
    >>> test = idx[int(idx.size / 2):]
    >>> model = ClassifierAdapter(SVC(probability=True))
    >>> nc = ClassifierNc(model, MarginErrFunc())
    >>> tcp = TcpClassifier(nc)
    >>> tcp.fit(iris.data[train, :], iris.target[train])
    >>> tcp.predict(iris.data[test, :], significance=0.10)
    ...             # doctest: +SKIP
    array([[ True, False, False],
            [False,  True, False],
            ...,
            [False,  True, False],
            [False,  True, False]], dtype=bool)
    """

    def __init__(self, nc_function, condition=None, smoothing=True):
        self.train_x, self.train_y = None, None
        self.nc_function = nc_function
        super(TcpClassifier, self).__init__()

        # Check if condition-parameter is the default function (i.e.,
        # lambda x: 0). This is so we can safely clone the object without
        # the clone accidentally having self.conditional = True.
        def default_condition(x):
            return 0

        is_default = callable(condition) and (
            condition.__code__.co_code == default_condition.__code__.co_code
        )

        if is_default:
            self.condition = condition
            self.conditional = False
        elif callable(condition):
            self.condition = condition
            self.conditional = True
        else:
            self.condition = lambda x: 0
            self.conditional = False

        self.smoothing = smoothing

        self.base_icp = IcpClassifier(
            self.nc_function, self.condition, self.smoothing
        )

        self.classes = None

    def fit(self, x, y):
        self.train_x, self.train_y = x, y
        self.classes = np.unique(y)

    def predict(self, x, significance=None):
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        significance : float or None
                Significance level (maximum allowed error rate) of predictions.
                Should be a float between 0 and 1. If ``None``, then the p-values
                are output rather than the predictions.

        Returns
        -------
        p : numpy array of shape [n_samples, n_classes]
                If significance is ``None``, then p contains the p-values for each
                sample-class pair; if significance is a float between 0 and 1, then
                p is a boolean array denoting which labels are included in the
                prediction sets.
        """
        n_test = x.shape[0]
        n_train = self.train_x.shape[0]
        p = np.zeros((n_test, self.classes.size))
        for i in range(n_test):
            for j, y in enumerate(self.classes):
                train_x = np.vstack([self.train_x, x[i, :]])
                train_y = np.hstack([self.train_y, y])
                self.base_icp.fit(train_x, train_y)
                scores = self.base_icp.nc_function.score(train_x, train_y)
                ngt = (scores[:-1] > scores[-1]).sum()
                neq = (scores[:-1] == scores[-1]).sum()

                p[i, j] = calc_p(n_train, ngt, neq, self.smoothing)

        if significance is not None:
            return p > significance
        else:
            return p

    def predict_conf(self, x):
        """Predict the output values for a set of input patterns, using
        the confidence-and-credibility output scheme.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
                Inputs of patters for which to predict output values.

        Returns
        -------
        p : numpy array of shape [n_samples, 3]
                p contains three columns: the first column contains the most
                likely class for each test pattern; the second column contains
                the confidence in the predicted class label, and the third column
                contains the credibility of the prediction.
        """
        p = self.predict(x, significance=None)
        label = p.argmax(axis=1)
        credibility = p.max(axis=1)
        for i, idx in enumerate(label):
            p[i, idx] = -np.inf
        confidence = 1 - p.max(axis=1)

        return np.array([label, confidence, credibility]).T



import numpy as np
import platform
import warnings
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from . import _boosterc as boosterc
from ..predictioninterval import PredictionInterval


class LSBoostRegressor(BaseEstimator, RegressorMixin):
    """LSBoost regressor.

    Attributes:

        n_estimators: int
            number of boosting iterations.

        learning_rate: float
            controls the learning speed at training time.

        n_hidden_features: int
            number of nodes in successive hidden layers.

        reg_lambda: float
            L2 regularization parameter for successive errors in the optimizer
            (at training time).

        row_sample: float
            percentage of rows chosen from the training set.

        col_sample: float
            percentage of columns chosen from the training set.

        dropout: float
            percentage of nodes dropped from the training set.

        tolerance: float
            controls early stopping in gradient descent (at training time).

        direct_link: bool
            indicates whether the original features are included (True) in model's
            fitting or not (False).

        verbose: int
            progress bar (yes = 1) or not (no = 0) (currently).

        seed: int
            reproducibility seed for nodes_sim=='uniform', clustering and dropout.

        backend: str
            type of backend; must be in ('cpu', 'gpu', 'tpu')

        solver: str
            type of 'weak' learner; currently in ('ridge', 'lasso')

        activation: str
            activation function: currently 'relu', 'relu6', 'sigmoid', 'tanh'

        type_pi: str.
            type of prediction interval; currently "kde" (default) or "bootstrap".
            Used only in `self.predict`, for `self.replications` > 0 and `self.kernel`
            in ('gaussian', 'tophat'). Default is `None`.

        replications: int.
            number of replications (if needed) for predictive simulation.
            Used only in `self.predict`, for `self.kernel` in ('gaussian',
            'tophat') and `self.type_pi = 'kde'`. Default is `None`.


    """

    def __init__(
        self,
        n_estimators=100,
        learning_rate=0.1,
        n_hidden_features=5,
        reg_lambda=0.1,
        row_sample=1,
        col_sample=1,
        dropout=0,
        tolerance=1e-4,
        direct_link=1,
        verbose=1,
        seed=123,
        backend="cpu",
        solver="ridge",
        activation="relu",
        type_pi=None,
        replications=None,
        kernel=None,
    ):
        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "`backend` must be in ('cpu', 'gpu', 'tpu')"

        assert solver in (
            "ridge",
            "lasso",
        ), "`solver` must be in ('ridge', 'lasso')"

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.n_hidden_features = n_hidden_features
        self.reg_lambda = reg_lambda
        self.row_sample = row_sample
        self.col_sample = col_sample
        self.dropout = dropout
        self.tolerance = tolerance
        self.direct_link = direct_link
        self.verbose = verbose
        self.seed = seed
        self.backend = backend
        self.obj = None
        self.solver = solver
        self.activation = activation
        self.type_pi = type_pi
        self.replications = replications
        self.kernel = kernel

    def fit(self, X, y, **kwargs):
        """Fit Booster (regressor) to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
               Target values.

            **kwargs: additional parameters to be passed to self.cook_training_set.

        Returns:

            self: object.
        """

        self.obj = boosterc.fit_booster_regressor(
            X=np.asarray(X, order="C"),
            y=np.asarray(y, order="C"),
            n_estimators=self.n_estimators,
            learning_rate=self.learning_rate,
            n_hidden_features=self.n_hidden_features,
            reg_lambda=self.reg_lambda,
            row_sample=self.row_sample,
            col_sample=self.col_sample,
            dropout=self.dropout,
            tolerance=self.tolerance,
            direct_link=self.direct_link,
            verbose=self.verbose,
            seed=self.seed,
            backend=self.backend,
            solver=self.solver,
            activation=self.activation,
        )

        self.n_estimators = self.obj["n_estimators"]

        self.X_ = X

        self.y_ = y

        return self

    def predict(self, X, level=95, method=None, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            level: int
                Level of confidence (default = 95)

            method: str
                `None`, or 'splitconformal', 'localconformal'
                prediction (if you specify `return_pi = True`)

            **kwargs: additional parameters to be passed to
                self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}
        """

        if "return_pi" in kwargs:
            assert method in (
                "splitconformal",
                "localconformal",
            ), "method must be in ('splitconformal', 'localconformal')"
            self.pi = PredictionInterval(
                obj=self,
                method=method,
                level=level,
                type_pi=self.type_pi,
                replications=self.replications,
                kernel=self.kernel,
            )
            self.pi.fit(self.X_, self.y_)
            self.X_ = None
            self.y_ = None
            preds = self.pi.predict(X, return_pi=True)
            return preds

        return boosterc.predict_booster_regressor(
            self.obj, np.asarray(X, order="C")
        )



#!/usr/bin/env python

"""
docstring
"""

# Authors: Henrik Linusson
# Yaniv Romano modified np.py file to include CQR
# T. Moudiki modified __init__.py to import classes

# __version__ = '2.1.0'

from .nc import (
    AbsErrorErrFunc,
    QuantileRegErrFunc,
    RegressorNc,
    RegressorNormalizer,
)
from .cp import IcpRegressor
from .base import RegressorAdapter

__all__ = [
    "AbsErrorErrFunc",
    "QuantileRegErrFunc",
    "RegressorAdapter",
    "RegressorNc",
    "RegressorNormalizer",
    "IcpRegressor",
]



import pandas as pd
import requests


def download(
    pkgname="MASS",
    dataset="Boston",
    source="https://cran.r-universe.dev/",
    **kwargs
):
    URL = source + pkgname + "/data/" + dataset + "/json"
    res = requests.get(URL)
    return pd.DataFrame(res.json(), **kwargs)



from .sampling.rowsubsampling import subsample
from .misc.misc import merge_two_dicts, flatten, is_float, is_factor
from .progress_bar import Progbar

__all__ = [
    "subsample",
    "merge_two_dicts",
    "flatten",
    "is_float",
    "is_factor",
    "Progbar",
]



import numpy as np
import pandas as pd
import pickle
import querier as qr
from tqdm import tqdm


def corrtarget_encoder(df, target, rho=0.4, verbose=1, seed=123):
    """Encode non-numerical columns using correlations.

    Parameters
    ----------
    df: a data frame
        a data frame

    target: str
        target column a.k.a response

    rho: float
        correlation between pseudo target (used for averaging) and target

    verbose: int
        currently 0 = nothing printed; 1 = progress bar printed

    seed: int
        reproducibility seed

    Returns
    --------
    a tuple: numerical data frame and achieved correlation

    """

    target_ = df[target].values
    target_mean = target_.mean()
    target_std = target_.std()
    n_target = len(target_)

    C = np.eye(2)
    C[0, 1] = rho
    C[1, 0] = rho
    C_ = np.linalg.cholesky(C).T

    np.random.seed(seed)
    temp = np.vstack((target_, np.random.normal(size=n_target))).T
    df_ = pickle.loads(pickle.dumps(df, -1))
    df_["pseudo_target"] = target_mean + np.dot(temp, C_)[:, 1] * target_std

    covariates_names = df.columns.values[df.columns.values != target].tolist()
    X = qr.select(df, ", ".join(covariates_names))
    X_dtypes = X.dtypes
    X_numeric = pickle.loads(pickle.dumps(df, -1))

    col_iterator = covariates_names if verbose == 0 else tqdm(covariates_names)

    for col in col_iterator:
        if X_dtypes[col] == np.object:  # something like a character string
            X_temp = qr.summarize(
                df_, req=col + ", avg(pseudo_target)", group_by=col
            )
            levels = np.unique(qr.select(X, col).values)

            for l in levels:
                qrobj = qr.Querier(X_temp)

                val = np.float(
                    qrobj.filtr(col + '== "' + l + '"')
                    .select("avg_pseudo_target")
                    .df.values
                )

                X_numeric = qr.setwhere(
                    X_numeric, col=col, val=l, replace=val, copy=False
                )

        else:  # a numeric column
            X_numeric[col] = X[col]

    return (
        X_numeric,
        np.corrcoef(df[target].values, df_["pseudo_target"].values)[0, 1],
    )
